@@ -8,13 +8,16 @@
 
 #include "util/common.h"
 
-#include "arm.h"
+#include "arm/arm.h"
+#include "core/log.h"
+
+mLOG_DECLARE_CATEGORY(GBA_BIOS);
 
 void GBASwi16(struct ARMCore* cpu, int immediate);
 void GBASwi32(struct ARMCore* cpu, int immediate);
 
 uint32_t GBAChecksum(uint32_t* memory, size_t size);
-const uint32_t GBA_BIOS_CHECKSUM;
-const uint32_t GBA_DS_BIOS_CHECKSUM;
+extern const uint32_t GBA_BIOS_CHECKSUM;
+extern const uint32_t GBA_DS_BIOS_CHECKSUM;
 
 #endif
