@@ -11,17 +11,20 @@
 #elif defined(__APPLE__)
 #include <OpenGL/gl.h>
 #else
+#ifdef _MSC_VER
+#include <windows.h>
+#endif
 #include <GL/gl.h>
 #endif
 
 #include "platform/video-backend.h"
 
-struct GBAGLContext {
+struct mGLContext {
 	struct VideoBackend d;
 
 	GLuint tex;
 };
 
-void GBAGLContextCreate(struct GBAGLContext*);
+void mGLContextCreate(struct mGLContext*);
 
 #endif
