@@ -19,10 +19,10 @@ Q_OBJECT
 public:
 	AudioProcessorQt(QObject* parent = nullptr);
 
-	virtual void setInput(GBAThread* input) override;
 	virtual unsigned sampleRate() const override;
 
 public slots:
+	virtual void setInput(mCoreThread* input) override;
 	virtual bool start() override;
 	virtual void pause() override;
 
