@@ -13,7 +13,7 @@
 #include "ui_GIFView.h"
 
 extern "C" {
-#include "platform/imagemagick/imagemagick-gif-encoder.h"
+#include "feature/imagemagick/imagemagick-gif-encoder.h"
 }
 
 namespace QGBA {
@@ -25,14 +25,14 @@ public:
 	GIFView(QWidget* parent = nullptr);
 	virtual ~GIFView();
 
-	GBAAVStream* getStream() { return &m_encoder.d; }
+	mAVStream* getStream() { return &m_encoder.d; }
 
 public slots:
 	void startRecording();
 	void stopRecording();
 
 signals:
-	void recordingStarted(GBAAVStream*);
+	void recordingStarted(mAVStream*);
 	void recordingStopped();
 
 private slots:
