@@ -14,7 +14,7 @@ struct Patch {
 	struct VFile* vf;
 
 	size_t (*outputSize)(struct Patch* patch, size_t inSize);
-	bool (*applyPatch)(struct Patch* patch, void* in, size_t inSize, void* out, size_t outSize);
+	bool (*applyPatch)(struct Patch* patch, const void* in, size_t inSize, void* out, size_t outSize);
 };
 
 bool loadPatch(struct VFile* vf, struct Patch* patch);
