@@ -18,7 +18,9 @@ char* strdup(const char* str);
 #endif
 
 char* strnrstr(const char* restrict s1, const char* restrict s2, size_t len);
+bool endswith(const char* restrict s1, const char* restrict end);
 
+size_t toUtf8(uint32_t unichar, char* buffer);
 int utfcmp(const uint16_t* utf16, const char* utf8, size_t utf16Length, size_t utf8Length);
 char* utf16to8(const uint16_t* utf16, size_t length);
 uint32_t utf8Char(const char** unicode, size_t* length);
@@ -26,7 +28,12 @@ uint32_t utf16Char(const uint16_t** unicode, size_t* length);
 
 int hexDigit(char digit);
 const char* hex32(const char* line, uint32_t* out);
+const char* hex24(const char* line, uint32_t* out);
 const char* hex16(const char* line, uint16_t* out);
+const char* hex12(const char* line, uint16_t* out);
 const char* hex8(const char* line, uint8_t* out);
+const char* hex4(const char* line, uint8_t* out);
+
+void rtrim(char* string);
 
 #endif
